/*****************************************************************************
 * File:            fm10k-dump.c
 * Creation Date:   December 16, 2020
 * Description:     Utility to dump Non Volatile Memory on SPI for FM10K devices. Requires FM10K kernel module running with UIO enabled.
 * Compile:         gcc -o fm10k-dump fm10k-dump.c
 * Usage:           ./fm10k-dump </dev/uio0> <output.bin> [readSizeInMegabits=8]
 *
 * Copyright (c) 2014 - 2015, Intel Corporation
 * Copyright (c) 2020, FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the copyright holders nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>

#define FM10K_UIO_SIZE 0x0000000004000000
#define FM10K_UIO_OFFSET 0

//From Datasheet, 11.25.1 MGMT Map, Table 11-37
#define FM10K_REGISTER_BASE 0

#define FM10K_REGISTER_SPI_TX_DATA ((0xC26) + (FM10K_REGISTER_BASE))
#define FM10K_REGISTER_SPI_RX_DATA ((0xC27) + (FM10K_REGISTER_BASE))
#define FM10K_REGISTER_SPI_HEADER ((0xC28) + (FM10K_REGISTER_BASE))
#define FM10K_REGISTER_SPI_CTRL   ((0xC29) + (FM10K_REGISTER_BASE))

#define FM10K_SPI_FREQ_KHZ 50000

#define FM10K_SPI_HEADER_COMMAND_READ_BYTES 0x3
#define FM10K_SPI_HEADER_COMMAND_READ_MANUFACTURER_INFORMATION 0x9F

/* Test for GCC >= 4.4.0, see https://gcc.gnu.org/gcc-4.4/changes.html : Packed bit-fields of type char were not properly bit-packed on many targets prior to GCC 4.4 */
#if __GNUC__ > 4 || (__GNUC__ == 4 && (__GNUC_MINOR__ >= 4)) || __clang__

typedef union {
    struct {
        uint16_t Freq : 10; // Actual speed is PCIE_REFCLK/(2*(1+Freq))
        uint8_t Enable : 1;
        uint8_t Command : 4; // 0b00 = reset
        uint8_t HeaderSize : 2; // 0b00 = 4 bytes
        uint8_t DataSize : 2; // 0b00 = 4 bytes
        uint8_t DataShiftMethod : 2; // 0b00 = SINGLE, 0b01 = DUAL, 0b10 = QUAD, 0b11 = Reserved
        uint8_t Busy : 1; // volatile ReadOnly
        uint8_t Selected : 1; // volatile ReadOnly
        uint8_t DirIO2 : 1; // 0b0 = Input, 0b1 = Output
        uint8_t DirIO3 : 1; // 0b0 = Input, 0b1 = Output
        uint8_t PinIO2 : 1; // volatile ReadWrite
        uint8_t PinIO3 : 1; // volatile ReadWrite
        uint8_t Reserved : 5;
    } __attribute__((packed)) fields;
    uint32_t value;
} SPI_CTRL;

typedef union {
    struct {
        uint8_t RESERVED : 8;
        uint8_t productVariant : 5;
        uint8_t subCode : 3;
        uint8_t densityCode : 5;
        uint8_t familyCode : 3;
        uint8_t manufacturerId : 8;
    } __attribute__((packed)) fields;
    uint32_t value;
} SPI_COMMAND_READ_MANUFACTURER_RESULT;

#else
#error "Packed bit-fields of type char were not properly bit-packed on many targets prior to GCC 4.4, upgrade to GCC >= 4.4."
#endif

void ReadRegister32(uintptr_t mem, uint32_t addr, uint32_t* value) {
    *value = *(((uint32_t*)mem) + addr);
}

void WriteRegister32(uintptr_t mem, uint32_t addr, uint32_t value) {
    *(uint32_t volatile*)(((uint32_t*)mem) + addr) = value;

    //If returning instantly this will caused missed writes. nanosleep wait of 1 nsec was too much. Busy wait works
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-but-set-variable"
    volatile uint32_t dummy;
    for(uint32_t i = 0; i < 100; ++i){
        dummy = i;
    }
#pragma GCC diagnostic pop
}

uint32_t get_interval_diff(struct timeval *begin, struct timeval *end) {
    struct timeval endT;
    struct timeval diff;

    if (end == NULL) {
        gettimeofday(&endT, NULL);
    } else {
        endT.tv_sec = end->tv_sec;
        endT.tv_usec = end->tv_usec;
    }

    diff.tv_sec  = endT.tv_sec - begin->tv_sec;
    diff.tv_usec = endT.tv_usec - begin->tv_usec;

    return (diff.tv_sec * 1000 + diff.tv_usec / 1000);
}

uint32_t fm10k_uio_spi_SetCtrlReg(uintptr_t mem, SPI_CTRL value){
    struct timeval startTime;
    uint8_t isTimeout = 0;
    SPI_CTRL spiCtrl = {0};

    WriteRegister32(mem, FM10K_REGISTER_SPI_CTRL, value.value);

    gettimeofday(&startTime, NULL);
    do{
        if(isTimeout){
            printf("Timeout waiting for SPI_CTRL.Busy 0x%02x\n", spiCtrl.value);
            return 1;
        }

        if(get_interval_diff(&startTime, NULL) > 50){
            isTimeout = 1;
        }

        ReadRegister32(mem, FM10K_REGISTER_SPI_CTRL, &spiCtrl.value);

    } while (spiCtrl.fields.Busy);

    /* write back SPI_CTRL with command = 0 */
    spiCtrl.fields.Command = 0;
    WriteRegister32(mem, FM10K_REGISTER_SPI_CTRL, spiCtrl.value);

    return 0;
}

void fm10k_uio_spi_Enable(uintptr_t mem){
    //ENABLE SPI
    SPI_CTRL spiCtrl;
    ReadRegister32(mem, FM10K_REGISTER_SPI_CTRL, &spiCtrl.value);
    /* keep current freq setting and set SPI Enable */
    SPI_CTRL newSpiCtrl;
    newSpiCtrl.value = 0;
    newSpiCtrl.fields.Freq = spiCtrl.fields.Freq;
    newSpiCtrl.fields.Enable = 1;
    WriteRegister32(mem, FM10K_REGISTER_SPI_CTRL, newSpiCtrl.value);
}

void fm10k_uio_spi_Disable(uintptr_t mem){
    //DISABLE SPI
    SPI_CTRL spiCtrl;
    ReadRegister32(mem, FM10K_REGISTER_SPI_CTRL, &spiCtrl.value);
    /* keep current freq setting and set SPI Enable = 0 */
    SPI_CTRL newSpiCtrl;
    newSpiCtrl.value = 0;
    newSpiCtrl.fields.Freq = spiCtrl.fields.Freq;
    WriteRegister32(mem, FM10K_REGISTER_SPI_CTRL, newSpiCtrl.value);
}


//Error if result.value is 0x00
SPI_COMMAND_READ_MANUFACTURER_RESULT fm10k_uio_spi_ReadManufacturerData(uintptr_t mem){

    SPI_COMMAND_READ_MANUFACTURER_RESULT result = {0};
    SPI_CTRL currentSpiCtrl;
    SPI_CTRL spiCtrl = {0};

    uint32_t header;
    int32_t freq;

    fm10k_uio_spi_Enable(mem);


    ReadRegister32(mem, FM10K_REGISTER_SPI_CTRL, &currentSpiCtrl.value);

    if(FM10K_SPI_FREQ_KHZ > 0){
        freq = ((100000 / (int)FM10K_SPI_FREQ_KHZ) / 2) - 1;
        if (freq < 0){
            freq = 0;
        }
        spiCtrl.fields.Freq = freq;
    }

    spiCtrl.fields.Enable = currentSpiCtrl.fields.Enable;

    /* header: command (1 byte: READ_BYTES) */
    header = FM10K_SPI_HEADER_COMMAND_READ_MANUFACTURER_INFORMATION;
    WriteRegister32(mem, FM10K_REGISTER_SPI_HEADER, header);

    /* first loop only: set send header flag. Following loops: shift only data. */
    spiCtrl.fields.Command |= 0b0001;
    spiCtrl.fields.Command |= 0b0100;
    spiCtrl.fields.HeaderSize = 1;
    spiCtrl.fields.DataSize = 4 & 0b11;

    if(fm10k_uio_spi_SetCtrlReg(mem, spiCtrl)){
        return result;
    }

    /* get data */
    ReadRegister32(mem, FM10K_REGISTER_SPI_RX_DATA, &result.value);

    /* release CS */
    spiCtrl.fields.Command = 0b1000;
    if(fm10k_uio_spi_SetCtrlReg(mem, spiCtrl)){
        result.value = 0;
        return result;
    }


    fm10k_uio_spi_Disable(mem);

    return result;
}


uint32_t fm10k_uio_spi_ReadFlash(uintptr_t mem, uint32_t address, uint8_t* data, int32_t len){

    SPI_CTRL currentSpiCtrl;
    SPI_CTRL spiCtrl = {0};

    uint32_t rxData;
    uint32_t header;
    int32_t freq;
    int32_t cnt;
    int32_t numRead;

    fm10k_uio_spi_Enable(mem);


    ReadRegister32(mem, FM10K_REGISTER_SPI_CTRL, &currentSpiCtrl.value);

    if(FM10K_SPI_FREQ_KHZ > 0){
        freq = ((100000 / (int)FM10K_SPI_FREQ_KHZ) / 2) - 1;
        if (freq < 0){
            freq = 0;
        }
        spiCtrl.fields.Freq = freq;
    }

    spiCtrl.fields.Enable = currentSpiCtrl.fields.Enable;

    /* header: command (1 byte: READ_BYTES) + address (3 bytes) */
    header = (FM10K_SPI_HEADER_COMMAND_READ_BYTES << 24) | (address & 0xffffff);
    WriteRegister32(mem, FM10K_REGISTER_SPI_HEADER, header);

    /* first loop only: set send header flag. Following loops: shift only data. */
    spiCtrl.fields.Command |= 0b0001;

    cnt = 0;
    while (cnt < len){
        /* determine the number of data bytes to read [1..4] */
        numRead = (len - cnt) > 3 ? 4 : (len - cnt);
        /* set 'shift data' flag and number of data bytes */
        spiCtrl.fields.Command |= 0b0100;
        spiCtrl.fields.DataSize = numRead & 0b11;

        if(fm10k_uio_spi_SetCtrlReg(mem, spiCtrl)){
            return 1;
        }

        /* get data */
        ReadRegister32(mem, FM10K_REGISTER_SPI_RX_DATA, &rxData);

        /* push the read data into the array */
        while (numRead) {
            numRead--;
            data[cnt++] = (rxData >> (numRead * 8)) & 0xff;
        }

        spiCtrl.fields.Command = 0b0000;
        spiCtrl.fields.DataSize = 0b0000;
    }

    /* release CS */
    spiCtrl.fields.Command = 0b1000;
    if(fm10k_uio_spi_SetCtrlReg(mem, spiCtrl)){
        return 1;
    }


    fm10k_uio_spi_Disable(mem);

    return 0;
}

int main(int argc, char** argv){
    int fdUIO;
    FILE* fdOutput;

    if(argc < 3){
        printf("Usage: %s </dev/uio0> <output.bin> [forceSizeInMegabits, default 8]\n", argv[0]);
        return 1;
    }

    fdOutput = fopen(argv[2], "wb");

    if(fdOutput == NULL){
        printf("Could not create %s\n", argv[2]);
        return 1;
    }

    fdUIO = open(argv[1], O_RDWR);
    if(fdUIO <= 0){
        printf("Unable to open uio device %s to read NVM\n", argv[1]);
        return 1;
    }
    void* memmapAddr;
    memmapAddr = mmap(NULL, FM10K_UIO_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fdUIO, 0);
    if (memmapAddr == MAP_FAILED) {
        printf("Unable to map uio device %s to read NVM\n", argv[1]);
        close(fdUIO);
        return 1;
    }

    SPI_COMMAND_READ_MANUFACTURER_RESULT manufacturerInfo = fm10k_uio_spi_ReadManufacturerData((uintptr_t)memmapAddr);
    if(manufacturerInfo.value == 0){
        printf("Error getting manufacturer information\n");
        return 1;
    }
    printf("Manufacturer Info 0x%08x :: JEDEC Manufacturer ID 0x%02x family 0x%02x density 0x%02x %02x:%02x\n", manufacturerInfo.value, manufacturerInfo.fields.manufacturerId, manufacturerInfo.fields.familyCode, manufacturerInfo.fields.densityCode, manufacturerInfo.fields.subCode, manufacturerInfo.fields.productVariant);

    unsigned int mbitsToRead = 8;


    switch (manufacturerInfo.value >> 8) {
        case 0x1f2701:
            printf("Device: Adesto AT45Dxxx Standard:1 32Mbit\n");
            mbitsToRead = 32;
            break;
        default:
            printf("Device: Unknown, defaulting %uMbit size\n", mbitsToRead);
            break;
    }

    if(argc > 3){
        if(mbitsToRead != 8){
            printf("WARNING: %uMbit size was already auto-detected\n", mbitsToRead);
        }
        mbitsToRead = strtoul(argv[3], NULL, 10);
        printf("Forcing %uMbit size\n", mbitsToRead);
    }

    if(mbitsToRead > 1024){
        printf("Too many MBit: %u\n", mbitsToRead);
        return 1;
    }

    if(mbitsToRead < 8){
        printf("Too few MBit: %u\n", mbitsToRead);
        return 1;
    }

    uint32_t bootImageSize = (mbitsToRead * 1024 * 1024) / 8;

    uint8_t* value = malloc(bootImageSize);

    uint32_t strideSize = 1024;

    for(uint32_t addr = 0; addr < bootImageSize; addr += strideSize){
        printf("\rread @ 0x%08x / 0x%08x %d bytes", addr, bootImageSize, strideSize);
        if(fm10k_uio_spi_ReadFlash((uintptr_t)memmapAddr, addr, value + addr, strideSize)){
            return 2;
        }
    }


    fwrite(value, 1, bootImageSize, fdOutput);

    free(value);

    fclose(fdOutput);

    printf("\nRead %u bytes\n", bootImageSize);

    return 0;
}